from builtins import str
import ast
import inspect
import os

from sentrylogger import sentry_constants as sntry_cnsts
from sentrylogger.log_packages import extra_logging_packages as log_packs

walked = False
funclist = []


def func_validity(frame, event, arg, log_funcs):
    """
    A helper function that checks if the event is a function call and needs to
    be logged to Sentry.

    Checks if the function called is a user defined function, or if the package
    is from the list of modules specified in the extra_logging_packages list.

    Parameters
    ----------
    frame: frame
        The stack frame from the code being run. Consists all the data associated to
        the function such as arguments, name, etc.
    event: String
        A string naming the type of notification. Example: call, return, etc.
    arg: Object
        An event specific value such as the value being returned, exception or None.

    Returns
    ----------
    Tuple or None: (message, data) or None
        message is the message to log at Sentry and data is extra data to be logged.
        data is a dictionary that contains information related to function call
        arguments.
        The function returns None if the event is not a call or if the function
        is not to be logged.
    """
    if event == "call":
        fname = frame.f_code.co_name
        if fname and inspect.getmodule(frame):
            if ((fname in log_funcs) or (inspect.getmodule(frame).__name__ in log_packs)):
                    args, _, _, values = inspect.getargvalues(frame)
                    data = {i: values[i] for i in args}
                    return (sntry_cnsts.FUNCTION_CALL + str(fname), data)
    return None


def top_level_funcs(body):
    """
    Returns functions from the given code body.

    Parses and looks for functions in the given body of code.

    Parameters
    ----------
    body: String
        The body of code.

    Returns
    ----------
    Tuple: (functions)
        Tuple of set of functions in the code.
    """
    return (f for f in body if isinstance(f, ast.FunctionDef))


def parse_ast(filename):
    """
    Returns the parse tree of a given code file.

    Parses the code after opening a file in 'rt' mode and generates a parse tree
    to return.

    Parameters
    ----------
    filename: String
        Name of the file to be parsed.

    Returns
    ----------
    String
        Parse tree of the given file.
    """
    with open(filename, "rt") as file:
        try:
            return ast.parse(file.read(), filename=filename)
        except (SyntaxError,ValueError) as error:
            pass


def fetch_all_funcs():
    """
    Adds all the functions present in '.py' files of a given directory and it's
    subdirectories to the list of functions to be logged at DEBUG level on
    Sentry.

    Walks in the Current Working Directory(cwd) and checks for '.py' files.
    Parses the files and looks for functions present in them; and finally adds
    them to the list of functions to be logged at DEBUG level.
    """
    global walked
    global funclist
    if walked:
        return funclist
    for root, dirs, files in os.walk(os.getcwd()):
        for filee in files:
            extension = os.path.splitext(filee)[-1]
            if ((extension != sntry_cnsts.FILE_EXT_PYC) and
                    (extension == sntry_cnsts.FILE_EXT_PY)):
                    tree = parse_ast(os.path.join(root, filee))
                    if tree is not None:
                        for func in top_level_funcs(tree.body):
                                    if func and func.name:
                                            funclist.append(func.name)
    return funclist
